@@ -1,40 +1,49 @@
 from __future__ import annotations
 
-from abc import ABCMeta, abstractmethod
-import argparse
 import copy
+import importlib.util
 import json
 import logging
+import os
 import random
 import statistics
-from typing import (
-    IO,
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    Dict,
-    Optional,
-    Type,
-    cast,
-    Iterator,
-    List,
-)
-
-import pkg_resources
+from abc import ABCMeta
+from abc import abstractmethod
+from typing import IO
+from typing import TYPE_CHECKING
+from typing import Any
+from typing import Callable
+from typing import Dict
+from typing import Iterator
+from typing import List
+from typing import Optional
+from typing import Type
 
 import keyring
 from jira import JIRA
 from rich.console import Console
 from rich.progress import TaskID
+from safdie import BaseCommand as SafdieBaseCommand
+from safdie import get_entrypoints
 from urllib3 import disable_warnings
 
 from .constants import APP_NAME
+from .constants import FORMATTER_ENTRYPOINT
+from .constants import FUNCTION_ENTRYPOINT
+from .constants import SOURCE_ENTRYPOINT
 from .exceptions import ConfigurationError
-from .types import ConfigDict, InstanceDefinition, SelectFieldDefinition, SchemaRow
-from .utils import save_config, get_functions_for_module
+from .types import ConfigDict
+from .types import InstanceDefinition
+from .types import SchemaRow
+from .types import SelectFieldDefinition
+from .utils import get_custom_function_dir
+from .utils import get_functions_for_module
+from .utils import save_config
 
 if TYPE_CHECKING:
-    from .query import Executor, CounterChannel, Query
+    from .query import CounterChannel
+    from .query import Executor
+    from .query import Query
 
 logger = logging.getLogger(__name__)
 
@@ -87,87 +96,59 @@ BUILTIN_FUNCTIONS: Dict[str, Callable] = {
         ],
     ),
     # Random
-    **get_functions_for_module(random, ["random", "randrange", "randint", "choice"],),
+    **get_functions_for_module(
+        random,
+        ["random", "randrange", "randint", "choice"],
+    ),
     # JSON
     "json_loads": json.loads,
-    "json_dumps": json.dumps,
 }
 REGISTERED_FUNCTIONS: Dict[str, Callable] = {}
 
 
-def get_installed_commands() -> Dict[str, Type[BaseCommand]]:
-    possible_commands: Dict[str, Type[BaseCommand]] = {}
-    for entry_point in pkg_resources.iter_entry_points(group="jira_select.commands"):
-        try:
-            loaded_class = entry_point.load()
-        except ImportError:
-            logger.warning(
-                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
-                entry_point,
-            )
-            continue
-        if not issubclass(loaded_class, BaseCommand):
-            logger.warning(
-                "Loaded entrypoint %s, but loaded class is "
-                "not a subclass of `jira_select.plugin.BaseCommand`.",
-                entry_point,
-            )
-            continue
-        possible_commands[entry_point.name] = loaded_class
-
-    return possible_commands
-
-
-class BaseCommand(metaclass=ABCMeta):
+class BaseCommand(SafdieBaseCommand):
     _jira: Optional[JIRA] = None
 
-    def __init__(self, config: ConfigDict, options: argparse.Namespace):
+    def __init__(self, *, config: ConfigDict, **kwargs):
         self._config: ConfigDict = config
-        self._options: argparse.Namespace = options
         self._console = Console(highlight=False)
-        super().__init__()
+        super().__init__(**kwargs)
 
     @property
     def config(self) -> ConfigDict:
-        """ Provides the configuration dictionary."""
+        """Provides the configuration dictionary."""
         return self._config
 
     def save_config(self) -> None:
-        """ Saves the existing configuration dictionary."""
+        """Saves the existing configuration dictionary."""
         save_config(self.config, self.options.config)
 
-    @property
-    def options(self) -> argparse.Namespace:
-        """ Provides options provided at the command-line."""
-        return self._options
-
     @property
     def console(self) -> Console:
-        """ Provides access to the console (see `rich.console.Console`."""
+        """Provides access to the console (see `rich.console.Console`."""
         return self._console
 
     @property
     def jira(self) -> JIRA:
-        """ Provides access to the configured Jira instance."""
+        """Provides access to the configured Jira instance."""
         if self._jira is None:
-            instance: Dict[InstanceDefinition] = cast(  # type: ignore
-                InstanceDefinition,
-                self.config.get("instances", {}).get(self.options.instance_name, {}),
+            instance = self.config.instances.get(
+                self.options.instance_name, InstanceDefinition()
             )
 
-            instance_url = self.options.instance_url or instance.get("url")
+            instance_url = self.options.instance_url or instance.url
             if not instance_url:
                 raise ConfigurationError(
                     "instance_url not set; please run `jira-select configure`."
                 )
 
-            username = self.options.username or instance.get("username")
+            username = self.options.username or instance.username
             if not username:
                 raise ConfigurationError(
                     "username not set; please run `jira-select configure`."
                 )
 
-            password = self.options.password or instance.get("password")
+            password = self.options.password or instance.password
             if not password:
                 password = keyring.get_password(APP_NAME, instance_url + username)
                 if not password:
@@ -178,9 +159,7 @@ class BaseCommand(metaclass=ABCMeta):
                         "`jira-select configure`."
                     )
 
-            verify = self.options.disable_certificate_verification or instance.get(
-                "verify"
-            )
+            verify = self.options.disable_certificate_verification or instance.verify
             if verify is None:
                 verify = True
             if verify is False:
@@ -197,47 +176,13 @@ class BaseCommand(metaclass=ABCMeta):
 
         return self._jira
 
-    @classmethod
-    def get_help(cls) -> str:
-        """ Retuurns help text for this function."""
-        return ""
-
-    @classmethod
-    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
-        """ Allows adding additional command-line arguments. """
-        pass
-
-    @abstractmethod
-    def handle(self) -> None:
-        """ This is where the work of your function starts. """
-        ...
-
 
 def get_installed_formatters() -> Dict[str, Type[BaseFormatter]]:
-    possible_formatters: Dict[str, Type[BaseFormatter]] = {}
-    for entry_point in pkg_resources.iter_entry_points(group="jira_select.formatters"):
-        try:
-            loaded_class = entry_point.load()
-        except ImportError:
-            logger.warning(
-                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
-                entry_point,
-            )
-            continue
-        if not issubclass(loaded_class, BaseFormatter):
-            logger.warning(
-                "Loaded entrypoint %s, but loaded class is "
-                "not a subclass of `jira_select.plugin.BaseFormatter`.",
-                entry_point,
-            )
-            continue
-        possible_formatters[entry_point.name] = loaded_class
-
-    return possible_formatters
+    return get_entrypoints(FORMATTER_ENTRYPOINT, BaseFormatter)
 
 
 class BaseFormatter(metaclass=ABCMeta):
-    def __init__(self, executor: Executor, stream: IO[str]):
+    def __init__(self, executor: Executor, stream: IO[bytes]):
         self._executor = executor
         self._stream = stream
 
@@ -273,32 +218,47 @@ class BaseFormatter(metaclass=ABCMeta):
         ...
 
 
-def register_function(name: str, fn: Callable):
-    """ Register a callable to be a function available in queries."""
-    REGISTERED_FUNCTIONS[name] = fn
+def register_function(fn: Callable):
+    """Register a callable to be a function available in queries."""
+    REGISTERED_FUNCTIONS[fn.__name__] = fn
 
 
 def get_installed_functions(jira: JIRA = None) -> Dict[str, Callable]:
     possible_commands: Dict[str, Callable] = copy.copy(BUILTIN_FUNCTIONS)
+
+    # Import any modules in the custom functions directory; as a
+    # side-effect of this, the functions will become listed within
+    # REGISTERED_FUNCTIONS
+    function_dir = get_custom_function_dir()
+    for dirname, subdirlist, filelist in os.walk(function_dir):
+        for filename in filelist:
+            if filename.endswith(".py"):
+                module_path_parts = ["user_scripts"]
+                if dirname != function_dir:
+                    module_path_parts.append(
+                        dirname[len(function_dir) + 1 :]
+                        .replace("/", ".")
+                        .replace("\\", ".")
+                    )
+                module_path_parts.append(os.path.splitext(filename)[0])
+                module_path = ".".join(module_path_parts)
+
+                full_path = os.path.join(function_dir, dirname, filename)
+                try:
+                    spec = importlib.util.spec_from_file_location(
+                        module_path, full_path
+                    )
+                    if not spec:
+                        continue
+                    user_module = importlib.util.module_from_spec(spec)
+                    spec.loader.exec_module(user_module)  # type: ignore
+                except Exception as e:
+                    logger.error("Could not import user script at %s: %s", full_path, e)
+
     possible_commands.update(REGISTERED_FUNCTIONS)
 
-    for entry_point in pkg_resources.iter_entry_points(group="jira_select.functions"):
-        try:
-            loaded_class = entry_point.load()
-        except ImportError:
-            logger.warning(
-                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
-                entry_point,
-            )
-            continue
-        if not issubclass(loaded_class, BaseFunction):
-            logger.warning(
-                "Loaded entrypoint %s, but loaded class is "
-                "not a subclass of `jira_select.plugin.BaseFunction`.",
-                entry_point,
-            )
-            continue
-        possible_commands[entry_point.name] = cast(Callable, loaded_class(jira))
+    for fn_name, fn in get_entrypoints(FUNCTION_ENTRYPOINT, BaseFunction).items():
+        possible_commands[fn_name] = fn(jira)
 
     return possible_commands
 
@@ -319,26 +279,7 @@ class BaseFunction(metaclass=ABCMeta):
 
 
 def get_installed_sources() -> Dict[str, Type[BaseSource]]:
-    possible_sources: Dict[str, Type[BaseSource]] = {}
-    for entry_point in pkg_resources.iter_entry_points(group="jira_select.sources"):
-        try:
-            loaded_class = entry_point.load()
-        except ImportError:
-            logger.warning(
-                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
-                entry_point,
-            )
-            continue
-        if not issubclass(loaded_class, BaseSource):
-            logger.warning(
-                "Loaded entrypoint %s, but loaded class is "
-                "not a subclass of `jira_select.plugin.BaseSource`.",
-                entry_point,
-            )
-            continue
-        possible_sources[entry_point.name] = loaded_class
-
-    return possible_sources
+    return get_entrypoints(SOURCE_ENTRYPOINT, BaseSource)
 
 
 class BaseSource(metaclass=ABCMeta):
@@ -361,7 +302,10 @@ class BaseSource(metaclass=ABCMeta):
 
         for entry in cls.get_schema(jira):
             fields.append(
-                {"expression": entry["id"], "column": entry["id"],}
+                SelectFieldDefinition(
+                    expression=entry.id,
+                    column=entry.id,
+                )
             )
 
         return fields
